SystemOrganization addCategory: #'Memcached-Client'!
SystemOrganization addCategory: #'Memcached-Commands'!
SystemOrganization addCategory: #'Memcached-Core'!
SystemOrganization addCategory: #'Memcached-Exceptions'!

Object subclass: #MDCacheClientBase
	instanceVariableNames: 'servers serverList lock'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Memcached-Client'!

MDCacheClientBase class
	instanceVariableNames: 'serverPool'!

MDCacheClientBase class
	instanceVariableNames: 'serverPool'!

MDCacheClientBase subclass: #MDCacheClient
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Memcached-Client'!

!MDCacheClient commentStamp: '<historical>' prior: 0!
I am the basic memcache client for memcache 1.2.6 or less.!

!MDCacheClient class methodsFor: 'defaults' stamp: 'RamonLeon 10/6/2010 22:11'!
defaultStaleTime
	^ 5 seconds! !

!MDCacheClient methodsFor: 'protocol' stamp: 'RamonLeon 10/6/2010 22:59'!
get: aString
	"when a get goes stale, reset the value with a stale flag to expire shortly, this prevents
	other gets from all stampeeding the database for fresh data all at once"

	| result needsRefreshed |
	result := super get: aString.
	result ifNil: [ ^ nil ].
	needsRefreshed := result second.
	(result first < DateAndTime now and: [ needsRefreshed not ])
		ifTrue: [ 
			"this 0 seconds will actually be just the defaultStaleTimeout"
			self
				set: aString
				value: result third
				expires: 0 seconds
				refreshed: true.	
			"this return of nil should cause client to reset value if necessary, otherwise
			the timeout will expire shortly"
			^ nil ].
	^ result third! !

!MDCacheClient methodsFor: 'protocol' stamp: 'RamonLeon 10/4/2010 00:12'!
set: aString value: aValue expires: aDuration
	^ self
		set: aString
		value: aValue
		expires: aDuration
		refreshed: false! !

!MDCacheClient methodsFor: 'private' stamp: 'RamonLeon 10/6/2010 22:58'!
set: aString value: aValue expires: aDuration refreshed: isRefreshed
	"embed the actual expire time in the data and add an additional stale time to
	memcached timeout to prevent stampeed when key expires; get does the opposite"

	| timeout |
	timeout := aDuration
		ifNil: [ DateAndTime now + 10 days ]
		ifNotNil: [ DateAndTime now + aDuration ].
	super
		set: aString
		value: { timeout. 	isRefreshed. aValue } 
		expires:
			(aDuration
				ifNil: [ 10 days ]
				ifNotNil: [ aDuration + self class defaultStaleTime ]).
	^ aValue! !

!MDCacheClientBase class methodsFor: 'defaults' stamp: 'RamonLeon 10/7/2010 15:03'!
defaultConnectionLifetime
	^ 10 minutes! !

!MDCacheClientBase class methodsFor: 'defaults' stamp: 'RamonLeon 10/2/2010 11:24'!
defaultConnectionRetries
	^ 5! !

!MDCacheClientBase class methodsFor: 'defaults' stamp: 'RamonLeon 10/3/2010 18:30'!
defaultSerializer
	^ ReferenceStream! !

!MDCacheClientBase class methodsFor: 'instance creation' stamp: 'RamonLeon 10/6/2010 19:03'!
onServers: aList
	^ (self new)
		serverList: aList;
		yourself! !

!MDCacheClientBase class methodsFor: 'socket-pool' stamp: 'RamonLeon 10/7/2010 15:00'!
returnServer: aServer
	^ self serverPool nextPut: aServer! !

!MDCacheClientBase class methodsFor: 'socket-pool' stamp: 'RamonLeon 10/8/2010 11:45'!
serverAt: aKey ifAbsent: aBlock
	| found |
	found := self serverPool findFirst: [ :e | e first = aKey ].
	found 	ifNil: [ ^ {aKey. (DateAndTime now). (aBlock value)} ].
	DateAndTime now - found second > self defaultConnectionLifetime
		ifFalse: [ ^ found ].
	found third quit execute.
	^ self serverAt: aKey ifAbsent: aBlock! !

!MDCacheClientBase class methodsFor: 'socket-pool' stamp: 'RamonLeon 10/7/2010 14:08'!
serverPool
	^ serverPool ifNil: [ serverPool := SharedQueue new ]! !

!MDCacheClientBase methodsFor: 'private' stamp: 'RamonLeon 10/6/2010 20:57'!
assertServers
	serverList ifNil: [ Error signal: 'Memcache servers not configured!!' ]! !

!MDCacheClientBase methodsFor: 'private' stamp: 'RamonLeon 10/7/2010 15:35'!
close
	servers ifNil: [ ^ self ].
	(servers reject: [ :e | e isNil ]) do: [ :e | self class returnServer: e ].
	servers := nil! !

!MDCacheClientBase methodsFor: 'protocol' stamp: 'RamonLeon 10/7/2010 15:20'!
decrement: aString
	self assertServers.
	^ self
		whileConnected: [ 
			((self serverForKey: aString) decr)
				key: aString asByteArray;
				execute ]! !

!MDCacheClientBase methodsFor: 'protocol' stamp: 'RamonLeon 10/8/2010 13:32'!
get: aString
	self assertServers.
	^ [ [ | item |
	item := self
		whileConnected: [ 
			((self serverForKey: aString) get)
				key: aString asByteArray;
				execute ].
	item data
		ifNotNil: [ :value | self class defaultSerializer unStream: item data ] ]
		on: MDKeyNotFoundError
		do: [ nil ] ]
		on: Error
		do: [ nil ]! !

!MDCacheClientBase methodsFor: 'protocol' stamp: 'RamonLeon 10/2/2010 13:56'!
get: aString ifAbsentPut: aBlock
	self assertServers.
	^ (self get: aString) ifNil: [ self set: aString value: aBlock value ]! !

!MDCacheClientBase methodsFor: 'protocol' stamp: 'RamonLeon 10/2/2010 13:56'!
get: aString ifAbsentPut: aBlock expires: aDuration
	self assertServers.
	^ (self get: aString) ifNil: [ self set: aString value: aBlock value expires: aDuration ]! !

!MDCacheClientBase methodsFor: 'protocol' stamp: 'RamonLeon 10/7/2010 15:20'!
increment: aString
	self assertServers.
	^ [ 
	self
		whileConnected: [ 
			((self serverForKey: aString) incr)
				key: aString asByteArray;
				execute ] ]
		on: MDKeyNotFoundError
		do: [ :e | 
			self setCounter: aString value: 0.
			e retry ]! !

!MDCacheClientBase methodsFor: 'private' stamp: 'RamonLeon 10/7/2010 15:29'!
lock
	^ lock ifNil: [ lock := Monitor new ]! !

!MDCacheClientBase methodsFor: 'private' stamp: 'RamonLeon 10/6/2010 20:42'!
open
	self assertServers.
	servers := serverList collect: [ :e | self startServer: e key onPort: e value asNumber ]! !

!MDCacheClientBase methodsFor: 'protocol' stamp: 'RamonLeon 10/7/2010 15:20'!
remove: aString
	self assertServers.
	[ 
	self
		whileConnected: [ 
			((self serverForKey: aString) delete)
				key: aString asByteArray;
				execute ] ]
		on: Error
		do: [ :e | 
			e notifyLog.
			nil ]! !

!MDCacheClientBase methodsFor: 'private' stamp: 'RamonLeon 10/7/2010 15:29'!
serverForKey: aKey
	"hash the key to one of the servers in the pool"
	^ (servers at: aKey hash \\ servers size + 1) third! !

!MDCacheClientBase methodsFor: 'accessing' stamp: 'RamonLeon 10/2/2010 10:48'!
serverList: aList
	serverList := aList
		collect: [ :e | 
			| server |
			server := e splitOn: ':'.
			((server first splitOn: '.') collect: [ :n | n asNumber ]) -> server second ]! !

!MDCacheClientBase methodsFor: 'protocol' stamp: 'RamonLeon 10/3/2010 22:26'!
set: aString value: aValue
	^ self set: aString value: aValue expires: nil! !

!MDCacheClientBase methodsFor: 'protocol' stamp: 'RamonLeon 10/7/2010 15:20'!
set: aString value: aValue expires: aDuration
	| serialized |
	self assertServers.
	^ [ 
	serialized := self class defaultSerializer streamedRepresentationOf: aValue.
	self
		whileConnected: [ 
			| command |
			command := ((self serverForKey: aString) set)
				key: aString asByteArray;
				data: serialized;
				expiration: (aDuration ifNil: [ 0 seconds ]) asSeconds asString asByteArray;
				execute ].
	aValue ] on: Error do: [ aValue ]! !

!MDCacheClientBase methodsFor: 'protocol' stamp: 'RamonLeon 10/8/2010 13:33'!
setCounter: aString value: anInteger
	self assertServers.
	^ [ 
	self
		whileConnected: [ 
			((self serverForKey: aString) set)
				key: aString asByteArray;
				data: anInteger asString asByteArray;
				execute ].
	anInteger ]
		on: Error
		do: [ anInteger ]! !

!MDCacheClientBase methodsFor: 'private' stamp: 'RamonLeon 10/7/2010 14:44'!
startServer: anIp onPort: aPort
	| retries |
	retries := 0.
	^ self class
		serverAt: (anIp joinUsing: '.') , ':' , aPort asString
		ifAbsent: [ 
			[ 
			retries := retries + 1.
			(MDServer new)
				host: anIp asByteArray;
				port: aPort;
				openProtocol ]
				on: ConnectionTimedOut
				do: [ :error | 
					retries > self class defaultConnectionRetries
						ifTrue: [ error pass ].
					100 milliSeconds asDelay wait.
					error retry ] ]! !

!MDCacheClientBase methodsFor: 'private' stamp: 'RamonLeon 10/8/2010 13:33'!
whileConnected: aBlock
	| retried |
	^ self lock
		critical: [ 
			retried := false.
			self open.
			[aBlock
				on: MDConnectionClosedError , ConnectionClosed
				do: [ :e | 
					retried
						ifTrue: [ e pass ].
					servers := nil.
					retried := true.
					self open.
					e retry ]]
				ensure: [ self close ] ]! !

MDCacheClientBase subclass: #MDCacheDbClient
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Memcached-Client'!

!MDCacheDbClient commentStamp: '<historical>' prior: 0!
I am a memcache client for persistent databases like toyko tyrant or memcached db which implmentes the memcache protocol but doesn't support expiration.  This client adds client based expiration; keep in mind these db's don't supprt LRU cache expiration and won't shrink when something expires.!

!MDCacheDbClient methodsFor: 'protocol' stamp: 'RamonLeon 10/3/2010 22:22'!
get: aSymbol
	"memcachedb doesn't support expiration, so we implement expiration in the client so it can still be used as 
	an expiring cache.  The set:value: messages wrap all values with a valid expiration time which is read here"

	| result |
	result := super get: aSymbol.
	(result isNil or: [ result first < DateAndTime now ])
		ifTrue: [ ^ nil ].
	^ result second! !

!MDCacheDbClient methodsFor: 'protocol' stamp: 'RamonLeon 10/5/2010 23:34'!
set: aSymbol value: aValue expires: aDuration
	"Wrap value in association with an expiration time, still passing up duration in case server changes in future"

	super
		set: aSymbol
		value:
			{(aDuration ifNil: [ DateAndTime year: 2020 day: 1 ] ifNotNil: [ DateAndTime now + aDuration ]).
			aValue}
		expires: aDuration.
	^ aValue! !

Object subclass: #MDCommand
	instanceVariableNames: 'protocol'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Memcached-Commands'!

!MDCommand class methodsFor: 'instance creation' stamp: 'pmm 7/22/2007 22:04'!
on: aProtocol
	^self new setProtocol: aProtocol! !

!MDCommand methodsFor: 'private' stamp: 'pmm 7/22/2007 22:10'!
decodeResponse: aByteArray
	protocol checkServerError: aByteArray! !

!MDCommand methodsFor: 'actions' stamp: 'pmm 7/22/2007 22:10'!
execute
	^self decodeResponse: (protocol execute: self)! !

!MDCommand methodsFor: 'private' stamp: 'pmm 7/22/2007 22:10'!
executeOnProtocol
	self writeToProtocol.
	^protocol readResponse! !

!MDCommand methodsFor: 'private' stamp: 'pmm 7/22/2007 22:49'!
keyNotFound: aString
	MDKeyNotFoundError raiseSignal: 'key not found: ', aString! !

!MDCommand methodsFor: 'private' stamp: 'pmm 7/22/2007 22:04'!
setProtocol: aProtocol
	protocol := aProtocol! !

!MDCommand methodsFor: 'private' stamp: 'pmm 7/22/2007 22:02'!
writeToProtocol
	self subclassResponsibility! !

MDCommand subclass: #MDCountCommand
	instanceVariableNames: 'key value'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Memcached-Commands'!

!MDCountCommand commentStamp: 'pmm 7/21/2007 16:07' prior: 0!
A MDCountCommand is a command that changes the value of a variable on the server.

Commands "incr" and "decr" are used to change data for some item
in-place, incrementing or decrementing it. The data for the item is
treated as decimal representation of a 32-bit unsigned integer. If the
current data value does not conform to such a representation, the
commands behave as if the value were 0. Also, the item must already
exist for incr/decr to work; these commands won't pretend that a
non-existent key exists with value 0; instead, they will fail.

Instance Variables
	key:		<String>
	value:		<Integer>


- key
	is the key of the item the client wishes to change

- value
	is the amount by which the client wants to increase/decrease
	the item. It is a decimal representation of a 32-bit unsigned integer.!

!MDCountCommand methodsFor: 'private' stamp: 'pmm 7/22/2007 22:50'!
decodeResponse: aByteArray
	super decodeResponse: aByteArray.
	aByteArray = (ByteArray withAll: #(78 79 84 95 70 79 85 78 68)) "'NOT_FOUND' asByteArray"
		ifTrue: [ self keyNotFound ].
	^protocol decodeUnsignedInteger: aByteArray startingAt: 1! !

!MDCountCommand methodsFor: 'accessing' stamp: 'pmm 7/22/2007 23:05'!
key: aByteArray 
	key := aByteArray! !

!MDCountCommand methodsFor: 'private' stamp: 'pmm 7/22/2007 22:49'!
keyNotFound
	self keyNotFound: key asString! !

!MDCountCommand methodsFor: 'private' stamp: 'pmm 7/22/2007 23:06'!
setProtocol: aProtocol
	super setProtocol: aProtocol.
	value := ByteArray with:  49 "$1 asciiValue"! !

!MDCountCommand methodsFor: 'accessing' stamp: 'pmm 7/22/2007 23:03'!
value: aByteArray 
	value := aByteArray! !

!MDCountCommand methodsFor: 'private' stamp: 'pmm 7/22/2007 22:44'!
writeToProtocol
	protocol space.
	protocol nextPutAll: key.
	protocol space.
	protocol nextPutAll: value.
	protocol endCommand! !

MDCountCommand subclass: #MDDecrementCommand
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Memcached-Commands'!

!MDDecrementCommand commentStamp: 'pmm 7/21/2007 16:09' prior: 0!
A MDDecrementCommand is a command that decrements a variable on the server.

Note that underflow in the "decr" command is caught: if a client tries
to decrease the value below 0, the new value will be 0.

Note also that decrementing a number such that it loses length isn't
guaranteed to decrement its returned length.  The number MAY be
space-padded at the end, but this is purely an implementation
optimization, so you also shouldn't rely on that.!

!MDDecrementCommand methodsFor: 'private' stamp: 'pmm 7/22/2007 22:45'!
writeToProtocol
	"decr <key> <value>\r\n"
	protocol nextPutAll: (ByteArray with: 100 with: 101 with: 99 with: 114). "decr' asByteArray"
	super writeToProtocol! !

MDCountCommand subclass: #MDIncrementCommand
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Memcached-Commands'!

!MDIncrementCommand commentStamp: 'pmm 7/21/2007 16:08' prior: 0!
A MDIncrementCommand is a command that increments a variable on the server.

Overflow in the "incr" command is not checked.!

!MDIncrementCommand methodsFor: 'private' stamp: 'pmm 7/22/2007 22:46'!
writeToProtocol
	"incr <key> <value>\r\n"
	protocol nextPutAll: (ByteArray with: 105 with: 110 with: 99 with: 114). "'incr' asByteArray"
	super writeToProtocol! !

MDCommand subclass: #MDDeleteCommand
	instanceVariableNames: 'key time'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Memcached-Commands'!

!MDDeleteCommand commentStamp: 'pmm 7/22/2007 12:29' prior: 0!
A MDDeleteCommand is a command which deletes an item on the server.

Instance Variables
	key:	<String>
	time:	<ByteArray>


key
	- is the key of the item the client wishes the server to delete
	
time
	- is the amount of time in seconds (or Unix time until which)
	the client wishes the server to refuse "add" and "replace" commands
	with this key. For this amount of item, the item is put into a
	delete queue, which means that it won't possible to retrieve it by
	the "get" command, but "add" and "replace" command with this key
	will also fail (the "set" command will succeed, however). After the
	time passes, the item is finally deleted from server memory.

	The parameter <time> is optional, and, if absent, defaults to 0
	(which means that the item will be deleted immediately and further
	storage commands with this key will succeed).
!

!MDDeleteCommand methodsFor: 'private' stamp: 'pmm 7/22/2007 22:11'!
decodeResponse: aByteArray
	super decodeResponse: aByteArray.
	(aByteArray = (ByteArray withAll: #(68 69 76 69 84 69 68))) "'DELETED' asByteArray"
		ifTrue: [ ^true ].
	(aByteArray = (ByteArray withAll: #(78 79 84 95 70 79 85 78 68))) "'NOT_FOUND' asByteArray"
		ifTrue: [ ^false ].
	protocol protocolError! !

!MDDeleteCommand methodsFor: 'accessing' stamp: 'pmm 7/21/2007 15:55'!
key: aString 
	key := aString! !

!MDDeleteCommand methodsFor: 'private' stamp: 'pmm 7/22/2007 22:05'!
setProtocol: aProtocol
	super setProtocol: aProtocol.
	time := (ByteArray with: 0)! !

!MDDeleteCommand methodsFor: 'private' stamp: 'pmm 7/22/2007 22:10'!
writeToProtocol
	"delete <key> <time>\r\n"
	protocol nextPutAll: (ByteArray withAll: #(100 101 108 101 116 101)). "'delete' asByteArray"
	protocol space.
	protocol nextPutAll: key.
	protocol space.
	protocol nextPutAll: time.
	protocol endCommand! !

MDCommand subclass: #MDFlushCommand
	instanceVariableNames: 'delay'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Memcached-Commands'!

!MDFlushCommand commentStamp: 'pmm 7/21/2007 16:14' prior: 0!
"flush_all" is a command with an optional numeric argument. It always
succeeds, and the server sends "OK\r\n" in response. Its effect is to
invalidate all existing items immediately (by default) or after the
expiration specified.  After invalidation none of the items will be returned
in response to a retrieval command (unless it's stored again under the
same key *after* flush_all has invalidated the items). flush_all doesn't
actually free all the memory taken up by existing items; that will
happen gradually as new items are stored. The most precise definition
of what flush_all does is the following: it causes all items whose
update time is earlier than the time at which flush_all was set to be
executed to be ignored for retrieval purposes.

The intent of flush_all with a delay, was that in a setting where you
have a pool of memcached servers, and you need to flush all content,
you have the option of not resetting all memcached servers at the
same time (which could e.g. cause a spike in database load with all
clients suddenly needing to recreate content that would otherwise
have been found in the memcached daemon).

The delay option allows you to have them reset in e.g. 10 second
intervals (by passing 0 to the first, 10 to the second, 20 to the
third, etc. etc.).!

!MDFlushCommand methodsFor: 'private' stamp: 'pmm 7/23/2007 07:08'!
decodeResponse: aByteArray
	super decodeResponse: aByteArray.
	aByteArray = (ByteArray with: 79 with: 75)"'OK' asByteArray"
		ifFalse: [ protocol protocolError ].
	^true! !

!MDFlushCommand methodsFor: 'accessing' stamp: 'pmm 7/23/2007 07:06'!
delay: aByteArray 
	delay := aByteArray! !

!MDFlushCommand methodsFor: 'private' stamp: 'pmm 7/23/2007 07:06'!
setProtocol: aProtocol
	super setProtocol: aProtocol.
	delay := ByteArray with: 48 "$0 asciiValue"! !

!MDFlushCommand methodsFor: 'private' stamp: 'pmm 7/23/2007 07:02'!
writeToProtocol
	protocol nextPutAll: (ByteArray withAll: #(102 108 117 115 104 95 97 108 108))."'flush_all' asByteArray"
	protocol endCommand! !

MDCommand subclass: #MDMultilineCommand
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Memcached-Commands'!

!MDMultilineCommand methodsFor: 'private' stamp: 'pmm 7/23/2007 06:45'!
decodeLine: aByteArray
	self subclassResponsibility! !

!MDMultilineCommand methodsFor: 'private' stamp: 'pmm 7/23/2007 07:04'!
executeOnProtocol
	self writeToProtocol.
	^Array streamContents: [ :stream |
		| response end |
		response := protocol readResponse.
		end := ByteArray with: 69 with: 78 with: 68. "'END' asByteArray"
		[ response = end ] whileFalse: [
			protocol checkServerError: response.
			stream nextPut: (self decodeLine: response).
			response := protocol readResponse ] ]! !

MDMultilineCommand subclass: #MDRetrievalCommand
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Memcached-Commands'!

MDRetrievalCommand subclass: #MDMultipleRetrievalCommand
	instanceVariableNames: 'keys'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Memcached-Commands'!

!MDRetrievalCommand methodsFor: 'private' stamp: 'pmm 7/23/2007 21:00'!
decodeLine: aByteArray
	| header key flags bytes data codec end |
	header := aByteArray first: 6.
	(header = (ByteArray withAll: #(86 65 76 85 69 32))) "'VALUE ' asByteArray"
		ifFalse: [ protocol protocolError ].
	codec := protocol codec.
	key := codec decodeAsciiString: aByteArray startingAt: 7.
	flags := codec decodeUnsignedInteger: aByteArray startingAt: 7 + key size.
	end := codec firstNonNumericIn: aByteArray startingAt: 9 + key size.
	bytes := codec decodeUnsignedInteger: aByteArray startingAt: end + 1.
	data := protocol readResponse: bytes.
	^MDItem key: key flags: flags data: data! !

!MDRetrievalCommand methodsFor: 'private' stamp: 'pmm 7/22/2007 13:29'!
decodeResponse: anArray
	"nothing to do, already done"
	^anArray! !

!MDRetrievalCommand methodsFor: 'private' stamp: 'pmm 7/22/2007 22:11'!
writeToProtocol
	protocol nextPutAll: (ByteArray with: 103 with: 101 with: 116). "'get' asByteArray"
	protocol space! !

MDRetrievalCommand subclass: #MDSingleRetrievalCommand
	instanceVariableNames: 'key'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Memcached-Commands'!

!MDSingleRetrievalCommand methodsFor: 'private' stamp: 'pmm 7/22/2007 22:01'!
executeOnProtocol
	| result |
	result := super executeOnProtocol.
	result isEmpty ifTrue: [ self keyNotFound ].
	^result at: 1! !

!MDSingleRetrievalCommand methodsFor: 'accessing' stamp: 'pmm 7/22/2007 16:16'!
key: aString 
	key := aString! !

!MDSingleRetrievalCommand methodsFor: 'private' stamp: 'pmm 7/22/2007 22:49'!
keyNotFound
	self keyNotFound: key asString! !

!MDSingleRetrievalCommand methodsFor: 'private' stamp: 'pmm 7/22/2007 22:11'!
writeToProtocol
	"get <key>*\r\n"
	super writeToProtocol.
	protocol nextPutAll: key.
	protocol endCommand! !

MDMultilineCommand subclass: #MDStatisticsCommand
	instanceVariableNames: 'arguments'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Memcached-Commands'!

!MDStatisticsCommand commentStamp: 'pmm 7/21/2007 16:13' prior: 0!
A MDStatisticsCommand is a command that queries statistics from the server.

Instance Variables
	arguments:		<Collection>

arguments
	- xxxxx
The command "stats" is used to query the server about statistics it
maintains and other internal data. It has two forms. Without
arguments:

stats\r\n

it causes the server to output general-purpose statistics and
settings, documented below.  In the other form it has some arguments:

stats <args>\r\n

Depending on <args>, various internal data is sent by the server. The
kinds of arguments and the data sent are not documented in this vesion
of the protocol, and are subject to change for the convenience of
memcache developers.


General-purpose statistics
--------------------------

Upon receiving the "stats" command without arguments, the server sents
a number of lines which look like this:

STAT <name> <value>\r\n

The server terminates this list with the line

END\r\n

In each line of statistics, <name> is the name of this statistic, and
<value> is the data.  The following is the list of all names sent in
response to the "stats" command, together with the type of the value
sent for this name, and the meaning of the value.

In the type column below, "32u" means a 32-bit unsigned integer, "64u"
means a 64-bit unsigner integer. '32u:32u' means two 32-but unsigned
integers separated by a colon.


Name              Type     Meaning
----------------------------------
pid               32u      Process id of this server process
uptime            32u      Number of seconds this server has been running
time              32u      current UNIX time according to the server
version           string   Version string of this server
rusage_user       32u:32u  Accumulated user time for this process 
                           (seconds:microseconds)
rusage_system     32u:32u  Accumulated system time for this process 
                           (seconds:microseconds)
curr_items        32u      Current number of items stored by the server
total_items       32u      Total number of items stored by this server 
                           ever since it started
bytes             64u      Current number of bytes used by this server 
                           to store items
curr_connections  32u      Number of open connections
total_connections 32u      Total number of connections opened since 
                           the server started running
connection_structures 32u  Number of connection structures allocated 
                           by the server
cmd_get           64u      Cumulative number of retrieval requests
cmd_set           64u      Cumulative number of storage requests
get_hits          64u      Number of keys that have been requested and 
                           found present
get_misses        64u      Number of items that have been requested 
                           and not found
evictions         64u      Number of valid items removed from cache                                                                           
                           to free memory for new items                                                                                       
bytes_read        64u      Total number of bytes read by this server 
                           from network
bytes_written     64u      Total number of bytes sent by this server to 
                           network
limit_maxbytes    32u      Number of bytes this server is allowed to
                           use for storage. !

!MDStatisticsCommand methodsFor: 'accessing' stamp: 'pmm 7/23/2007 06:40'!
arguments: aByteArray 
	arguments := aByteArray! !

!MDStatisticsCommand methodsFor: 'private' stamp: 'pmm 7/23/2007 20:56'!
decodeLine: aByteArray
	| name value size |
	(aByteArray beginsWith: (ByteArray withAll: #(83 84 65 84))) "'STAT' asByteArray"
		ifFalse: [ protocol protocolError ].
	name := protocol codec decodeAsciiString: aByteArray startingAt: 6.
	size := aByteArray size.
	value := aByteArray copyFrom: name size + 7 to: size.
	^Array with: name with: value! !

!MDStatisticsCommand methodsFor: 'private' stamp: 'pmm 7/23/2007 23:24'!
decodeResponse: anArray
	"nothing to do, already done"
	| stats codec processId uptime time version userTime systemTime currentItems totalItems bytesUsed currentConnections totalConnections connectionStructures retrievalRequests storageRequests hits misses evictions bytesRead bytesWritten maximumBytes zero pointerSize |
	stats := Dictionary new.
	codec := protocol codec.
	anArray do: [ :each |
		stats at: (each at: 1) put: (each at: 2) ].
	zero := ByteArray with: 48.
	processId := codec decodeUnsignedInteger: (stats at: 'pid').
	uptime := codec seconds: (codec decodeUnsignedInteger: (stats at: 'uptime')).
	time := codec asDateAndTime: (codec decodeUnsignedInteger: (stats at: 'time')).
	version := codec decodeAsciiString: (stats at: 'version').
	userTime := codec decodeSecondsMilliseconds: (stats at: 'rusage_user').
	systemTime := codec decodeSecondsMilliseconds: (stats at: 'rusage_system').
	currentItems := codec decodeUnsignedInteger: (stats at: 'curr_items').
	totalItems := codec decodeUnsignedInteger: (stats at: 'total_items').
	bytesUsed := codec decodeUnsignedInteger: (stats at: 'bytes').
	currentConnections := codec decodeUnsignedInteger: (stats at: 'curr_connections').
	totalConnections := codec decodeUnsignedInteger: (stats at: 'total_connections').
	connectionStructures := codec decodeUnsignedInteger: (stats at: 'connection_structures').
	retrievalRequests := codec decodeUnsignedInteger: (stats at: 'cmd_get').
	storageRequests := codec decodeUnsignedInteger: (stats at: 'cmd_set').
	hits := codec decodeUnsignedInteger: (stats at: 'get_hits').
	misses := codec decodeUnsignedInteger: (stats at: 'get_misses').
	evictions := codec decodeUnsignedInteger: (stats at: 'evictions' ifAbsent: [ zero ]).
	bytesRead := codec decodeUnsignedInteger: (stats at: 'bytes_read').
	bytesWritten := codec decodeUnsignedInteger: (stats at: 'bytes_written').
	maximumBytes := codec decodeUnsignedInteger: (stats at: 'limit_maxbytes').
	pointerSize := codec decodeUnsignedInteger: (stats at: 'pointer_size' ifAbsent: [ zero ]).
	^MDStatistic new
		processId: processId;
		uptime: uptime;
		time: time;
		version: version;
		userTime: userTime;
		systemTime: systemTime;
		currentItems: currentItems;
		totalItems: totalItems;
		bytesUsed: bytesUsed;
		currentConnections: currentConnections;
		totalConnections: totalConnections;
		connectionStructures: connectionStructures;
		retrievalRequests: retrievalRequests;
		storageRequests: storageRequests;
		hits: hits;
		misses: misses;
		evictions: evictions;
		bytesRead: bytesRead;
		bytesWritten: bytesWritten;
		maximumBytes: maximumBytes;
		pointerSize: pointerSize;
		yourself! !

!MDStatisticsCommand methodsFor: 'private' stamp: 'pmm 7/23/2007 06:42'!
writeToProtocol
	protocol nextPutAll: (ByteArray withAll: # (115 116 97 116 115))."'stats' asByteArray"
	arguments isNil ifFalse: [
		protocol space.
		protocol nextPutAll: arguments ].
	protocol endCommand! !

MDCommand subclass: #MDQuitCommand
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Memcached-Commands'!

!MDQuitCommand commentStamp: 'pmm 7/21/2007 16:20' prior: 0!
A MDQuitCommand is a command that upon receiving causes the server to close the connection.

However, the client may also simply close the connection when it no longer needs it, without issuing this command.!

!MDQuitCommand methodsFor: 'private' stamp: 'pmm 7/22/2007 12:08'!
decodeResponse: aByteArray
	"no response will be sent, server will close upon receive"! !

!MDQuitCommand methodsFor: 'private' stamp: 'pmm 7/22/2007 22:02'!
executeOnProtocol
	self writeToProtocol
	"don't read response, there will be none"! !

!MDQuitCommand methodsFor: 'private' stamp: 'pmm 7/22/2007 22:10'!
writeToProtocol
	"quit\r\n"
	protocol nextPutAll: (ByteArray withAll: #(113 117 105 116)). "'quit' asByteArray"
	protocol endCommand.
	protocol closeStream! !

MDCommand subclass: #MDStorageCommand
	instanceVariableNames: 'name key flags expiration data bytes'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Memcached-Commands'!

!MDStorageCommand commentStamp: 'pmm 7/22/2007 11:56' prior: 0!
A MDStorageCommand is a command that stores data on the server.

Instance Variables
	name:		<String>
	key:		<String>
	flags:		<SmallInteger>
	expiration:	<Integer>
	data:		<ByteArray>
	bytes:		<ByteArray>


name
	- 'set', 'add' or 'replace'
	  "set" means "store this data".  

	  "add" means "store this data, but only if the server *doesn't* already
	  hold data for this key".  

	  "replace" means "store this data, but only if the server *does*
	  already hold data for this key".

key
	- the key under which the client asks to store the data. A key
	is a text string which should uniquely identify the data for clients
	that are interested in storing and retrieving it.  Currently the
	length limit of a key is set at 250 characters (of course, normally
	clients wouldn't need to use such long keys); the key must not include
	control characters or whitespace.

flags
	- an arbitrary 16-bit unsigned integer (written out in
	decimal) that the server stores along with the data and sends back
	 when the item is retrieved. Clients may use this as a bit field to
	store data-specific information; this field is opaque to the server.
	Note that in memcached 1.2.1 and higher, flags may be 32-bits, instead
	of 16, but you might want to restrict yourself to 16 bits for
	compatibility with older versions.

expirationTime
	- expiration time. If it's 0, the item never expires
	(although it may be deleted from the cache to make place for other
	items). If it's non-zero (either Unix time or offset in seconds from
	current time), it is guaranteed that clients will not be able to
	retrieve this item after the expiration time arrives (measured by
	server time). 

data
	- is a chunk of arbitrary 8-bit data
	
bytes
	 - is the number of bytes in the data block to follow, *not*
	including the delimiting \r\n. <bytes> may be zero (in which case
	it's followed by an empty data block).!

!MDStorageCommand methodsFor: 'actions' stamp: 'pmm 7/22/2007 11:42'!
add
	name :=  ByteArray with: 97 with: 100 with: 100 "'add' asByteArray"! !

!MDStorageCommand methodsFor: 'private' stamp: 'pmm 7/29/2007 16:34'!
argumentError: aString
	MDArgumentError raiseSignal: aString! !

!MDStorageCommand methodsFor: 'private' stamp: 'pmm 7/29/2007 16:34'!
data: aByteArray
	aByteArray size > 16r100000
		ifTrue: [ self argumentError: 'can not store more than 1 MiB' ].
	data := aByteArray.
	bytes := protocol codec encodeIntegerToByteArray: aByteArray size! !

!MDStorageCommand methodsFor: 'private' stamp: 'pmm 7/22/2007 22:11'!
decodeResponse: aByteArray
	super decodeResponse: aByteArray.
	(aByteArray = (ByteArray withAll: #(83 84 79 82 69 68))) "'STORED' asByteArray"
		ifTrue: [ ^true ].
	(aByteArray = (ByteArray withAll: #(78 79 84 95 83 84 79 82 69 68))) "'NOT_STORED' asByteArray"
		ifTrue: [ ^false ].
	protocol protocolError! !

!MDStorageCommand methodsFor: 'accessing' stamp: 'pmm 7/21/2007 15:57'!
expiration
	^expiration! !

!MDStorageCommand methodsFor: 'accessing' stamp: 'pmm 7/21/2007 15:58'!
expiration: anInteger 
	expiration := anInteger! !

!MDStorageCommand methodsFor: 'accessing' stamp: 'pmm 7/21/2007 15:58'!
flags: anInteger 
	flags := anInteger! !

!MDStorageCommand methodsFor: 'accessing' stamp: 'pmm 7/21/2007 15:58'!
key: anInteger 
	key := anInteger! !

!MDStorageCommand methodsFor: 'actions' stamp: 'pmm 7/22/2007 11:43'!
replace
	name :=  ByteArray withAll: #(114 101 112 108 97 99 101) "'replace' asByteArray"! !

!MDStorageCommand methodsFor: 'actions' stamp: 'pmm 7/22/2007 11:38'!
set
	name :=  ByteArray with: 115 with: 101 with: 116 "'set' asByteArray"! !

!MDStorageCommand methodsFor: 'private' stamp: 'pmm 7/22/2007 22:05'!
setProtocol: aProtocol
	super setProtocol: aProtocol.
	expiration := ByteArray with: 48. "$0 asciiValue"
	flags := ByteArray with: 48.  "$0 asciiValue"
	data := ByteArray new! !

!MDStorageCommand methodsFor: 'private' stamp: 'pmm 7/22/2007 22:11'!
writeToProtocol
	"<command name> <key> <flags> <exptime> <bytes>\r\n"
	protocol nextPutAll: name.
	protocol space.
	protocol nextPutAll: key.
	protocol space.
	protocol nextPutAll: flags.	
	protocol space.
	protocol nextPutAll: expiration.
	protocol space.
	protocol nextPutAll: bytes.
	protocol crlf.
	"<data block>\r\n"
	protocol nextPutAll: data.
	protocol endCommand! !

MDCommand subclass: #MDVerbosityCommand
	instanceVariableNames: 'level'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Memcached-Commands'!

!MDVerbosityCommand commentStamp: 'pmm 7/21/2007 16:17' prior: 0!
A MDVerbosityCommand is a command that sets the verbosity level of the logging output.

Instance Variables
	level:		<anInteger>

level
	- the verbosity level
!

MDCommand subclass: #MDVersionCommand
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Memcached-Commands'!

!MDVersionCommand commentStamp: 'pmm 7/21/2007 16:16' prior: 0!
A MDVersionCommand is a command the returns the the version string for the server.
!

!MDVersionCommand methodsFor: 'private' stamp: 'pmm 7/23/2007 20:56'!
decodeResponse: aByteArray
	super decodeResponse: aByteArray.
	"VERSION <version>"
	(aByteArray beginsWith: #(86 69 82 83 73 79 78)) "'VERSION' asByteArray"
		ifFalse: [ protocol protocolError ].
	^protocol codec decodeAsciiString: aByteArray startingAt: 9! !

!MDVersionCommand methodsFor: 'private' stamp: 'pmm 7/22/2007 22:09'!
writeToProtocol
	"version\r\n"
	protocol nextPutAll: (ByteArray withAll: #(118 101 114 115 105 111 110)). "'version' asByteArray"
	protocol endCommand! !

Object subclass: #MDInternalCodec
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Memcached-Core'!

MDInternalCodec subclass: #MDChronosCodec
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Memcached-Core'!

!MDChronosCodec methodsFor: 'converting' stamp: 'pmm 7/23/2007 19:17'!
asDateAndTime: anInteger
	^(ChronosSystemClock unixEpoch) + (ScientificDuration seconds: anInteger)! !

!MDChronosCodec methodsFor: 'converting' stamp: 'pmm 7/23/2007 19:09'!
asDuration: anInteger
	^ScientificDuration seconds: anInteger! !

!MDChronosCodec methodsFor: 'converting' stamp: 'pmm 7/23/2007 20:52'!
seconds: secondsInteger microseconds: microsecondsInteger
	^ScientificDuration seconds: secondsInteger nanoseconds: 1000 * microsecondsInteger! !

!MDInternalCodec methodsFor: 'private' stamp: 'pmm 7/23/2007 19:01'!
argumentError: aString
	MDArgumentError raiseSignal: aString! !

!MDInternalCodec methodsFor: 'converting' stamp: 'pmm 7/23/2007 19:15'!
asDateAndTime: anInteger
	^(DateAndTime year: 1970 month: 1 day: 1 hour: 0 minute: 0 second: 0 offset: (Duration seconds: 0)) + (Duration seconds: anInteger)! !

!MDInternalCodec methodsFor: 'converting' stamp: 'pmm 7/23/2007 19:08'!
asDuration: anInteger
	^Duration seconds: anInteger! !

!MDInternalCodec methodsFor: 'decoding' stamp: 'pmm 7/23/2007 22:58'!
decodeAsciiString: aByteArray
	^self decodeAsciiString: aByteArray startingAt: 1! !

!MDInternalCodec methodsFor: 'decoding' stamp: 'pmm 7/23/2007 18:51'!
decodeAsciiString: aByteArray startingAt: anInteger
	| buffer totalSize |
	totalSize := aByteArray size.
	buffer := WriteStream on: (String new: totalSize - anInteger).
	anInteger to: totalSize do: [ :index |
		| value |
		value := aByteArray at: index.
		value = 32 "space"
			ifTrue: [ ^buffer contents ].
		buffer nextPut: (Character value: value) ].
	^buffer contents! !

!MDInternalCodec methodsFor: 'decoding' stamp: 'pmm 7/23/2007 23:06'!
decodeSecondsMilliseconds: aByteArray
	| split |
	split := self firstNonNumericIn: aByteArray startingAt: 1.
	^self
		seconds: (self decodeUnsignedInteger: (aByteArray first: split - 1))
		microseconds: (self decodeUnsignedInteger: aByteArray startingAt: split + 1)! !

!MDInternalCodec methodsFor: 'decoding' stamp: 'pmm 7/23/2007 22:56'!
decodeUnsignedInteger: aByteArray
	^self decodeUnsignedInteger: aByteArray startingAt: 1! !

!MDInternalCodec methodsFor: 'decoding' stamp: 'pmm 7/23/2007 18:51'!
decodeUnsignedInteger: aByteArray startingAt: anInteger
	| buffer totalSize |
	totalSize := aByteArray size.
	buffer := 0.
	anInteger to: totalSize do: [ :index |
		| value |
		value := aByteArray at: index.
		value = 32 "space"
			ifTrue: [ ^buffer ].
		buffer := (buffer * 10) + (value - 48) ].
	^buffer! !

!MDInternalCodec methodsFor: 'encoding' stamp: 'pmm 7/23/2007 19:00'!
encodeIntegerToByteArray: anInteger
	| decimal bytes |
	decimal := anInteger printStringBase: 10.
	bytes := ByteArray new: decimal size.
	1 to: decimal size do: [ :index |
		bytes at: index put: (decimal at: index) asciiValue ].
	^bytes! !

!MDInternalCodec methodsFor: 'encoding' stamp: 'pmm 7/23/2007 19:03'!
encodeUInt32: anInteger
	anInteger < 0
		ifTrue: [ self argumentError: 'can only encode positive integers' ].
	anInteger >  16rFFFFFFFF
		ifTrue: [ self argumentError: 'can only encode integers smallte than: ', 16r100000000 printString ].
	^self encodeIntegerToByteArray: anInteger! !

!MDInternalCodec methodsFor: 'encoding' stamp: 'pmm 7/23/2007 19:05'!
encodeUInt64: anInteger
	anInteger < 0
		ifTrue: [ self argumentError: 'can only encode positive integers' ].
	anInteger >  16rFFFFFFFFFFFFFFFF
		ifTrue: [ self argumentError: 'can only encode integers smallte than: ', 16r10000000000000000 printString ].
	^self encodeIntegerToByteArray: anInteger! !

!MDInternalCodec methodsFor: 'decoding' stamp: 'pmm 7/23/2007 20:44'!
firstNonNumericIn: aByteArray startingAt: anInteger
	| index each |
	index := anInteger.
	each := aByteArray at: index.
	[ each >= 48 and: [ each <= 57 ] ] whileTrue: [
		index := index + 1.
		each := aByteArray at: index.
	].
	^index! !

!MDInternalCodec methodsFor: 'converting' stamp: 'pmm 7/23/2007 22:55'!
seconds: secondsInteger
	^self seconds: secondsInteger microseconds: 0! !

!MDInternalCodec methodsFor: 'converting' stamp: 'pmm 7/23/2007 19:19'!
seconds: secondsInteger microseconds: microsecondsInteger
	^Duration seconds: secondsInteger nanoSeconds: 1000 * microsecondsInteger! !

Object subclass: #MDItem
	instanceVariableNames: 'key flags data'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Memcached-Commands'!

!MDItem class methodsFor: 'instance creation' stamp: 'pmm 7/22/2007 13:20'!
key: aString flags: anInteger data: aByteArray
	^self new
		initializeWithKey: aString flags: anInteger data: aByteArray! !

!MDItem methodsFor: 'accessing' stamp: 'pmm 7/22/2007 13:19'!
data
	^data! !

!MDItem methodsFor: 'accessing' stamp: 'pmm 7/22/2007 13:19'!
data: anObject
	data := anObject! !

!MDItem methodsFor: 'accessing' stamp: 'pmm 7/22/2007 13:19'!
flags
	^flags! !

!MDItem methodsFor: 'accessing' stamp: 'pmm 7/22/2007 13:19'!
flags: anObject
	flags := anObject! !

!MDItem methodsFor: 'initialize-release' stamp: 'pmm 7/22/2007 13:20'!
initializeWithKey: aString flags: anInteger data: aByteArray
	key := aString.
	flags := anInteger.
	data := aByteArray.! !

!MDItem methodsFor: 'accessing' stamp: 'pmm 7/22/2007 13:19'!
key
	^key! !

!MDItem methodsFor: 'accessing' stamp: 'pmm 7/22/2007 13:19'!
key: anObject
	key := anObject! !

Object subclass: #MDNullCodec
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Memcached-Core'!

!MDNullCodec methodsFor: 'decoding' stamp: 'pmm 7/29/2007 10:57'!
decodeFloat: aByteArray
	^Float fromIEEE32Bit: aByteArray! !

!MDNullCodec methodsFor: 'decoding' stamp: 'pmm 7/31/2007 06:53'!
decodeObject: aByteArray
	^(ReferenceStream on: aByteArray readStream) next
	! !

!MDNullCodec methodsFor: 'decoding' stamp: 'pmm 7/29/2007 11:05'!
decodeString: aByteArray
	^aByteArray asString! !

!MDNullCodec methodsFor: 'encoding' stamp: 'pmm 7/29/2007 10:56'!
encodeFloat: aFloat
	^aFloat asIEEE32BitWord! !

!MDNullCodec methodsFor: 'encoding' stamp: 'pmm 7/30/2007 21:32'!
encodeInteger: anInteger
	"high bytes come first"
	| remaining stream |
	remaining := anInteger.
	"#byteSize"
	stream := WriteStream on: (ByteArray new: 8).
	[| next |
	next := remaining bitAnd: 16rFF.
	remaining := remaining bitShift: -8.
	(remaining > 0 and: [ remaining ~= -1 ])
		ifTrue: [
			stream nextPut: next.
			true ]
		ifFalse: [
			next > 16r7F
				ifTrue: [
					remaining = -1
						ifTrue: [
							stream nextPut: next ]
						ifFalse: [
							stream nextPut: (next bitAnd: 16r7F).
							stream nextPut: 16r01 ] ]
				ifFalse: [
					remaining = -1
						ifTrue: [
							stream nextPut: next.
							stream nextPut: 16rFF  ]
						ifFalse: [
							stream nextPut: next ] ].
			false ] ] whileTrue.
	^stream contents reversed! !

!MDNullCodec methodsFor: 'decoding' stamp: 'pmm 7/31/2007 06:51'!
encodeObject: anObject
	"^ByteArray streamContents: [ :stream |
		| segment srefStream |
		segment := ImageSegment new.
		segment copyFromRootsForExport: (Array with: anObject).
		srefStream := SmartRefStream on: (MultiByteBinaryOrTextStream on: stream).
		srefStream nextPut: segment. ]"	
	"^ByteArray streamContents: [ :stream |
		| srefStream |
		srefStream := SmartRefStream on: stream.
		srefStream nextPut: anObject.
		srefStream close ]"
	| stream referenceStream |
	stream := MultiByteBinaryOrTextStream on: (ByteArray new: 100).
	referenceStream := ReferenceStream on: stream.
	referenceStream nextPut: anObject.
	referenceStream close.
	^stream reset binary contents.
	"self basicNew encodeObject: OrderedCollection new"! !

!MDNullCodec methodsFor: 'encoding' stamp: 'pmm 7/29/2007 11:04'!
encodeString: aString
	^aString asByteArray! !

MDNullCodec subclass: #MDUtf8Codec
	instanceVariableNames: ''
	classVariableNames: 'Latin1ToUtf8Encodings Latin1ToUtf8Map'
	poolDictionaries: ''
	category: 'Memcached-Core'!

!MDUtf8Codec class methodsFor: 'class initialization' stamp: 'pmm 7/24/2007 07:05'!
initialize
	"MDUtf8Coded initialize"
	| latin1 utf8 |
	Latin1ToUtf8Map := ByteArray new: 256.
	Latin1ToUtf8Encodings := Array new: 256.
	0 to: 255 do:[ :index |
		latin1 := String with: (Character value: index).
		utf8 := latin1 convertToEncoding: #utf8.
		latin1 = utf8
			ifTrue: [
				Latin1ToUtf8Map at: index + 1 put: 0 ] "no translation needed"
			ifFalse: [
				Latin1ToUtf8Map at: index + 1 put: 1. "no translation needed"
				Latin1ToUtf8Encodings at: index + 1 put: utf8 ] ]! !

!MDUtf8Codec methodsFor: 'private' stamp: 'pmm 7/24/2007 07:10'!
invalidUtf8
	MDArgumentError signal: 'Invalid UTF-8 input'! !

!MDUtf8Codec methodsFor: 'private' stamp: 'pmm 7/24/2007 07:17'!
latin1ToUtf8: aByteString
	"Convert the given string from UTF-8 using the fast path if converting to Latin-1"
	| outStream lastIndex nextIndex |
	lastIndex := 1.
	nextIndex := ByteString findFirstInString: aByteString inSet: Latin1ToUtf8Map startingAt: lastIndex.
	nextIndex = 0 ifTrue:[ ^aByteString ].
	outStream := (String new: aByteString size * 2) writeStream.
	[	outStream next: nextIndex - lastIndex putAll: aByteString startingAt: lastIndex.
		outStream nextPutAll: (Latin1ToUtf8Encodings at: (aByteString byteAt: nextIndex) + 1).
		lastIndex := nextIndex + 1.
		nextIndex := ByteString findFirstInString: aByteString inSet: Latin1ToUtf8Map startingAt: lastIndex.
		nextIndex = 0 ] whileFalse.
	outStream next: aByteString size - lastIndex + 1 putAll: aByteString startingAt: lastIndex.
	^outStream contents! !

!MDUtf8Codec methodsFor: 'encoding' stamp: 'pmm 7/24/2007 07:16'!
toUtf8: aString
	^aString isByteString
		ifTrue: [ self latin1ToUtf8: aString ]
		ifFalse: [ aString convertToEncoding: #utf8 ]! !

!MDUtf8Codec methodsFor: 'decoding' stamp: 'pmm 7/24/2007 07:12'!
utf8ToLatin1: aByteString
	"Convert the given string from UTF-8 using the fast path if converting to Latin-1"
	| outStream lastIndex nextIndex byte1 byte2 byte3 byte4 unicode |
	lastIndex := 1.
	nextIndex := ByteString findFirstInString: aByteString inSet: Latin1ToUtf8Map startingAt: lastIndex.
	nextIndex = 0 ifTrue:[ ^aByteString ].
	outStream := (String new: aByteString size) writeStream.
		[outStream next: nextIndex-lastIndex putAll: aByteString startingAt: lastIndex.
		byte1 := aByteString byteAt: nextIndex.
		(byte1 bitAnd: 16rE0) = 192 ifTrue: [ "two bytes"
			byte2 := aByteString byteAt: (nextIndex := nextIndex+1).
			(byte2 bitAnd: 16rC0) = 16r80 ifFalse:[ self invalidUtf8 ].
			unicode := ((byte1 bitAnd: 31) bitShift: 6) + (byte2 bitAnd: 63)].
		(byte1 bitAnd: 16rF0) = 224 ifTrue: [ "three bytes"
			byte2 := aByteString byteAt: (nextIndex := nextIndex+1).
			(byte2 bitAnd: 16rC0) = 16r80 ifFalse:[ self invalidUtf8 ].
			byte3 := aByteString byteAt: (nextIndex := nextIndex+1).
			(byte3 bitAnd: 16rC0) = 16r80 ifFalse:[ self invalidUtf8 ].
			unicode := ((byte1 bitAnd: 15) bitShift: 12) + ((byte2 bitAnd: 63) bitShift: 6)
				+ (byte3 bitAnd: 63)].
		(byte1 bitAnd: 16rF8) = 240 ifTrue: [ "four bytes"
			byte2 := aByteString byteAt: (nextIndex := nextIndex+1).
			(byte2 bitAnd: 16rC0) = 16r80 ifFalse:[ self invalidUtf8 ].
			byte3 := aByteString byteAt: (nextIndex := nextIndex+1).
			(byte3 bitAnd: 16rC0) = 16r80 ifFalse:[ self invalidUtf8 ].
			byte4 := aByteString byteAt: (nextIndex := nextIndex+1).
			(byte4 bitAnd: 16rC0) = 16r80 ifFalse:[ self invalidUtf8 ].
			unicode := ((byte1 bitAnd: 16r7) bitShift: 18) +
							((byte2 bitAnd: 63) bitShift: 12) + 
							((byte3 bitAnd: 63) bitShift: 6) +
							(byte4 bitAnd: 63)].
		unicode ifNil:[ self invalidUtf8 ].
		outStream nextPut: (Character value: unicode).
		lastIndex := nextIndex + 1.
		nextIndex := ByteString findFirstInString: aByteString inSet: Latin1ToUtf8Map startingAt: lastIndex.
		nextIndex = 0 ] whileFalse.
	outStream next: aByteString size - lastIndex + 1 putAll: aByteString startingAt: lastIndex.
	^outStream contents! !

Object subclass: #MDProtocol
	instanceVariableNames: 'stream mutex codec'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Memcached-Core'!

!MDProtocol commentStamp: 'pmm 7/22/2007 22:00' prior: 0!
A MDProtocol is a low level implementation of the memcached protocol that deals with bytes.!

!MDProtocol class methodsFor: 'instance creation' stamp: 'pmm 7/22/2007 00:35'!
on: aStream
	^self new initializeWithStream: aStream! !

!MDProtocol methodsFor: 'commands' stamp: 'pmm 7/22/2007 11:49'!
add
	"'add' means 'store this data, but only if the server *doesn't* already
  hold data for this key'"
	^self storageCommand add! !

!MDProtocol methodsFor: 'private-decoding' stamp: 'pmm 7/22/2007 10:45'!
checkServerError: aByteArray
	(aByteArray beginsWith: #(69 82 82 79 82)) "'ERROR' asByteArray"
		ifTrue: [ self noSuchCommand ].
	(aByteArray beginsWith: #(67 76 73 69 78 84 95 69 82 82 79 82)) "'CLIENT_ERROR' asByteArray"
		ifTrue: [ self clientError: aByteArray ].
	(aByteArray beginsWith: #(83 69 82 86 69 82 95 69 82 82 79 82)) "'SERVER_ERROR' asByteArray"
		ifTrue: [ self clientError: aByteArray ]! !

!MDProtocol methodsFor: 'private-signaling' stamp: 'pmm 7/23/2007 20:53'!
clientError: aByteArray
	"CLIENT_ERROR <error>"
	MDClientError raiseSignal: (codec decodeAsciiString: aByteArray startingAt: 14)! !

!MDProtocol methodsFor: 'private' stamp: 'pmm 7/22/2007 10:56'!
closeStream
	stream close! !

!MDProtocol methodsFor: 'accessing' stamp: 'pmm 7/23/2007 19:22'!
codec
	^codec! !

!MDProtocol methodsFor: 'private' stamp: 'pmm 7/23/2007 19:21'!
codecClass
	^(Smalltalk hasClassNamed: #ChronosObject)
		ifTrue: [ MDChronosCodec ]
		ifFalse: [ MDInternalCodec ]! !

!MDProtocol methodsFor: 'private-commands' stamp: 'pmm 7/22/2007 00:41'!
command: aCommandClass
	^aCommandClass on: self! !

!MDProtocol methodsFor: 'private-signaling' stamp: 'RamonLeon 10/8/2010 12:54'!
connectionClosed
	MDConnectionClosedError raiseSignal! !

!MDProtocol methodsFor: 'writing' stamp: 'pmm 7/22/2007 12:06'!
crlf
	stream nextPut: 13.
	stream nextPut: 10.! !

!MDProtocol methodsFor: 'private-decoding' stamp: 'pmm 7/22/2007 13:02'!
decodeAsciiString: aByteArray startingAt: anInteger
	| buffer totalSize |
	totalSize := aByteArray size.
	buffer := WriteStream on: (String new: totalSize - anInteger).
	anInteger to: totalSize do: [ :index |
		| value |
		value := aByteArray at: index.
		value = 32 "space"
			ifTrue: [ ^buffer contents ].
		buffer nextPut: (Character value: value) ].
	^buffer contents! !

!MDProtocol methodsFor: 'private-decoding' stamp: 'pmm 7/22/2007 16:27'!
decodeUnsignedInteger: aByteArray startingAt: anInteger
	| buffer totalSize |
	totalSize := aByteArray size.
	buffer := 0.
	anInteger to: totalSize do: [ :index |
		| value |
		value := aByteArray at: index.
		value = 32 "space"
			ifTrue: [ ^buffer ].
		buffer := (buffer * 10) + (value - 48) ].
	^buffer! !

!MDProtocol methodsFor: 'commands' stamp: 'pmm 7/22/2007 22:51'!
decr
	^self command: MDDecrementCommand! !

!MDProtocol methodsFor: 'commands' stamp: 'pmm 7/22/2007 12:36'!
delete
	^self command: MDDeleteCommand! !

!MDProtocol methodsFor: 'writing' stamp: 'pmm 7/22/2007 12:06'!
endCommand
	self crlf.
	stream flush! !

!MDProtocol methodsFor: 'private' stamp: 'pmm 7/22/2007 22:01'!
execute: aCommand
	^mutex critical: [ aCommand executeOnProtocol ]! !

!MDProtocol methodsFor: 'commands' stamp: 'pmm 7/23/2007 07:08'!
flushAll
	^self command: MDFlushCommand! !

!MDProtocol methodsFor: 'commands' stamp: 'pmm 7/22/2007 13:31'!
get
	^self command: MDSingleRetrievalCommand! !

!MDProtocol methodsFor: 'commands' stamp: 'pmm 7/22/2007 22:51'!
incr
	^self command: MDIncrementCommand! !

!MDProtocol methodsFor: 'initialize-release' stamp: 'pmm 7/23/2007 19:22'!
initializeWithStream: aStream
	stream := aStream binary.
	mutex := Semaphore forMutualExclusion.
	codec := self codecClass new! !

!MDProtocol methodsFor: 'writing' stamp: 'pmm 7/22/2007 00:45'!
nextPutAll: aByteArray
	stream nextPutAll: aByteArray! !

!MDProtocol methodsFor: 'private-signaling' stamp: 'pmm 7/22/2007 10:39'!
noSuchCommand
	MDNoSuchCommandError raiseSignal! !

!MDProtocol methodsFor: 'private-signaling' stamp: 'pmm 7/22/2007 10:48'!
protocolError
	"SERVER_ERROR <error>"
	MDProtocolError raiseSignal! !

!MDProtocol methodsFor: 'commands' stamp: 'pmm 7/22/2007 10:56'!
quit
	^self command: MDQuitCommand! !

!MDProtocol methodsFor: 'reading' stamp: 'RamonLeon 10/8/2010 12:55'!
readResponse
	"read up to CR LF into a byte array"

	stream atEnd 	ifTrue: [ self connectionClosed ].
	^ByteArray streamContents: [ :buffer |
		| last secondLast |
		secondLast := stream next.
		last := stream next.
		[ secondLast = 13
			and: [ last = 10 ] ]
				whileFalse: [
					buffer nextPut: secondLast.
					secondLast := last.
					last := stream next ] ]! !

!MDProtocol methodsFor: 'reading' stamp: 'pmm 7/22/2007 16:20'!
readResponse: anInteger
	"read up to CR LF into a byte array"
	| data cr lf |
	data := stream next: anInteger.
	cr := stream next.
	lf := stream next.
	(cr ~= 13 or: [ lf ~= 10 ])
		ifTrue: [ self protocolError ].
	^data! !

!MDProtocol methodsFor: 'commands' stamp: 'pmm 7/22/2007 16:46'!
replace
	"'replace' means 'store this data, but only if the server *does*
  already hold data for this key'"
	^self storageCommand replace! !

!MDProtocol methodsFor: 'private-signaling' stamp: 'pmm 7/23/2007 20:54'!
serverError: aByteArray
	"SERVER_ERROR <error>"
	MDClientError raiseSignal: (codec decodeAsciiString: aByteArray startingAt: 14)! !

!MDProtocol methodsFor: 'commands' stamp: 'pmm 7/22/2007 11:49'!
set
	"'set' means 'store this data'"
	^self storageCommand set! !

!MDProtocol methodsFor: 'writing' stamp: 'pmm 7/22/2007 00:44'!
space
	stream nextPut: 32.! !

!MDProtocol methodsFor: 'commands' stamp: 'pmm 7/23/2007 06:53'!
stats
	^self command: MDStatisticsCommand! !

!MDProtocol methodsFor: 'private-commands' stamp: 'pmm 7/22/2007 11:48'!
storageCommand
	^self command: MDStorageCommand! !

!MDProtocol methodsFor: 'commands' stamp: 'pmm 7/22/2007 01:03'!
version
	^self command: MDVersionCommand! !

Object subclass: #MDServer
	instanceVariableNames: 'host port'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Memcached-Core'!

!MDServer class methodsFor: 'instance creation' stamp: 'pmm 7/22/2007 10:52'!
default
	^self new
		host: self defaultHost;
		port: self defaultPort;
		yourself! !

!MDServer class methodsFor: 'accessing-defaults' stamp: 'pmm 7/22/2007 10:51'!
defaultHost
	^#(127 0 0 1) asByteArray! !

!MDServer class methodsFor: 'accessing-defaults' stamp: 'pmm 7/22/2007 10:51'!
defaultPort
	^11211! !

!MDServer class methodsFor: 'instance creation' stamp: 'pmm 7/22/2007 10:52'!
host: aStringOrByteArray
	^self new
		host: aStringOrByteArray;
		port: self defaultPort;
		yourself! !

!MDServer class methodsFor: 'instance creation' stamp: 'pmm 7/22/2007 00:32'!
host: aStringOrByteArray port: anInteger
	^self new
		host: aStringOrByteArray;
		port: anInteger;
		yourself! !

!MDServer class methodsFor: 'instance creation' stamp: 'pmm 7/22/2007 10:52'!
port: anInteger
	^self new
		host: self defaultHost;
		port: anInteger;
		yourself! !

!MDServer methodsFor: 'accessing' stamp: 'pmm 7/22/2007 00:32'!
host
	^host! !

!MDServer methodsFor: 'accessing' stamp: 'pmm 7/22/2007 00:35'!
host: aStringOrByteArray
	"aStringOrByteArray can either be a String with the hostname or a ByteArray with the host address"
	host := aStringOrByteArray! !

!MDServer methodsFor: 'actions' stamp: 'pmm 7/22/2007 21:59'!
openProtocol
	| stream |
	stream := self host isString
		ifTrue: [ SocketStream openConnectionToHostNamed: self host port: self port ]
		ifFalse: [ SocketStream openConnectionToHost: self host port: self port ].
	^MDProtocol on: stream! !

!MDServer methodsFor: 'accessing' stamp: 'pmm 7/22/2007 00:32'!
port
	^port! !

!MDServer methodsFor: 'accessing' stamp: 'pmm 7/22/2007 00:32'!
port: anObject
	port := anObject! !

Object subclass: #MDSettings
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Memcached-Core'!

!MDSettings methodsFor: 'as yet unclassified' stamp: 'pmm 7/21/2007 15:59'!
noStringEncoding
	^false! !

!MDSettings methodsFor: 'as yet unclassified' stamp: 'pmm 7/21/2007 15:59'!
use32BitFlags
	^false! !

Object subclass: #MDStatistic
	instanceVariableNames: 'processId uptime time version userTime systemTime totalItems currentItems bytesUsed currentConnections totalConnections connectionStructures retrievalRequests storageRequests hits misses evictions bytesRead bytesWritten maximumBytes pointerSize'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Memcached-Core'!

!MDStatistic commentStamp: 'pmm 7/23/2007 23:24' prior: 0!
A MDStatistic is a memcached server statistic.

Instance Variables
	bytesRead:		<Integer>
	bytesUsed:		<Integer>
	bytesWritten:		<Integer>
	connectionStructures:	<Integer>
	currentConnections:		<Integer>
	currentItems:		<Integer>
	evictions:		<Integer>
	hits:		<Integer>
	maximumBytes:		<Integer>
	misses:		<Integer>
	pointerSize: 		<Integer>
	processId:		<Integer>
	retrievalRequests:		<Integer>
	storageRequests:		<Integer>
	systemTime:		<Duration>
	time:		<DateAndTime>
	toatlConnections:		<Integer>
	totalItems:		<Integer>
	uptime:		<Duration>
	userTime:		<Duration>
	version:		<String>

bytesRead
	- Total number of bytes read by this server from network

bytesUsed
	- Current number of bytes used by this server to store items

bytesWritten
	- Total number of bytes sent by this server to network

connectionStructures
	- Number of connection structures allocated by the server
	
currentConnections
	- Number of open connections

currentItems
	- Current number of items stored by the server

evictions
	- Number of valid items removed from cache to free memory for new items

hits
	- Number of keys that have been requested and found present

maximumBytes
	- Number of bytes this server is allowed to use for storage. 

misses
	- Number of items that have been requested and not found
	
pointerSize
	- pointer size in bits

processId
	- Process id of this server process

retrievalRequests
	- Cumulative number of retrieval requests

storageRequests
	- Cumulative number of storage requests

systemTime
	- Accumulated system time for this process
	
totalConnections
	- Total number of connections opened since the server started running

time
	- current UNIX time according to the server

totalItems
	- Total number of items stored by this server ever since it started

uptime
	- Number of seconds this server has been running

userTime
	- Accumulated user time for this process

version
	- Version string of this server
!

!MDStatistic methodsFor: 'accessing' stamp: 'pmm 7/23/2007 22:50'!
bytesRead
	^bytesRead! !

!MDStatistic methodsFor: 'accessing' stamp: 'pmm 7/23/2007 22:50'!
bytesRead: anInteger 
	bytesRead := anInteger! !

!MDStatistic methodsFor: 'accessing' stamp: 'pmm 7/23/2007 22:49'!
bytesUsed
	^bytesUsed! !

!MDStatistic methodsFor: 'accessing' stamp: 'pmm 7/23/2007 22:50'!
bytesUsed: anInteger 
	bytesUsed := anInteger! !

!MDStatistic methodsFor: 'accessing' stamp: 'pmm 7/23/2007 22:50'!
bytesWritten
	^bytesWritten! !

!MDStatistic methodsFor: 'accessing' stamp: 'pmm 7/23/2007 22:50'!
bytesWritten: anInteger 
	bytesWritten := anInteger! !

!MDStatistic methodsFor: 'accessing' stamp: 'pmm 7/23/2007 22:49'!
connectionStructures
	^connectionStructures! !

!MDStatistic methodsFor: 'accessing' stamp: 'pmm 7/23/2007 22:50'!
connectionStructures: anInteger 
	connectionStructures := anInteger! !

!MDStatistic methodsFor: 'accessing' stamp: 'pmm 7/23/2007 22:49'!
currentConnections
	^currentConnections! !

!MDStatistic methodsFor: 'accessing' stamp: 'pmm 7/23/2007 22:51'!
currentConnections: anInteger 
	currentConnections := anInteger! !

!MDStatistic methodsFor: 'accessing' stamp: 'pmm 7/23/2007 22:49'!
currentItems
	^currentItems! !

!MDStatistic methodsFor: 'accessing' stamp: 'pmm 7/23/2007 22:51'!
currentItems: anInteger 
	currentItems := anInteger! !

!MDStatistic methodsFor: 'accessing' stamp: 'pmm 7/23/2007 22:50'!
evictions
	^evictions! !

!MDStatistic methodsFor: 'accessing' stamp: 'pmm 7/23/2007 22:51'!
evictions: anInteger 
	evictions := anInteger! !

!MDStatistic methodsFor: 'accessing' stamp: 'pmm 7/23/2007 22:49'!
hits
	^hits! !

!MDStatistic methodsFor: 'accessing' stamp: 'pmm 7/23/2007 22:51'!
hits: anInteger 
	hits := anInteger! !

!MDStatistic methodsFor: 'accessing' stamp: 'pmm 7/23/2007 22:50'!
maximumBytes
	^maximumBytes! !

!MDStatistic methodsFor: 'accessing' stamp: 'pmm 7/23/2007 22:51'!
maximumBytes: anInteger 
	maximumBytes := anInteger! !

!MDStatistic methodsFor: 'accessing' stamp: 'pmm 7/23/2007 22:50'!
misses
	^misses! !

!MDStatistic methodsFor: 'accessing' stamp: 'pmm 7/23/2007 22:51'!
misses: anInteger 
	misses := anInteger! !

!MDStatistic methodsFor: 'accessing' stamp: 'pmm 7/23/2007 23:21'!
pointerSize
	^pointerSize! !

!MDStatistic methodsFor: 'accessing' stamp: 'pmm 7/23/2007 23:21'!
pointerSize: anInteger 
	pointerSize := anInteger! !

!MDStatistic methodsFor: 'accessing' stamp: 'pmm 7/23/2007 22:49'!
processId
	^processId! !

!MDStatistic methodsFor: 'accessing' stamp: 'pmm 7/23/2007 22:51'!
processId: anInteger 
	processId := anInteger! !

!MDStatistic methodsFor: 'accessing' stamp: 'pmm 7/23/2007 22:49'!
retrievalRequests
	^retrievalRequests! !

!MDStatistic methodsFor: 'accessing' stamp: 'pmm 7/23/2007 22:51'!
retrievalRequests: anInteger 
	retrievalRequests := anInteger! !

!MDStatistic methodsFor: 'accessing' stamp: 'pmm 7/23/2007 22:49'!
storageRequests
	^storageRequests! !

!MDStatistic methodsFor: 'accessing' stamp: 'pmm 7/23/2007 22:52'!
storageRequests: anInteger 
	storageRequests := anInteger! !

!MDStatistic methodsFor: 'accessing' stamp: 'pmm 7/23/2007 22:49'!
systemTime
	^systemTime! !

!MDStatistic methodsFor: 'accessing' stamp: 'pmm 7/23/2007 22:52'!
systemTime: aDuration 
	systemTime := aDuration! !

!MDStatistic methodsFor: 'accessing' stamp: 'pmm 7/23/2007 22:49'!
time
	^time! !

!MDStatistic methodsFor: 'accessing' stamp: 'pmm 7/23/2007 22:52'!
time: aDateAndTime 
	time := aDateAndTime! !

!MDStatistic methodsFor: 'accessing' stamp: 'pmm 7/23/2007 22:49'!
totalConnections
	^totalConnections! !

!MDStatistic methodsFor: 'accessing' stamp: 'pmm 7/23/2007 22:52'!
totalConnections: anInteger 
	totalConnections := anInteger! !

!MDStatistic methodsFor: 'accessing' stamp: 'pmm 7/23/2007 22:49'!
totalItems
	^totalItems! !

!MDStatistic methodsFor: 'accessing' stamp: 'pmm 7/23/2007 22:53'!
totalItems: anInteger 
	totalItems := anInteger! !

!MDStatistic methodsFor: 'accessing' stamp: 'pmm 7/23/2007 22:49'!
uptime
	^uptime! !

!MDStatistic methodsFor: 'accessing' stamp: 'pmm 7/23/2007 22:53'!
uptime: aDuration 
	uptime := aDuration! !

!MDStatistic methodsFor: 'accessing' stamp: 'pmm 7/23/2007 22:49'!
userTime
	^userTime! !

!MDStatistic methodsFor: 'accessing' stamp: 'pmm 7/23/2007 22:53'!
userTime: aDuration 
	userTime := aDuration! !

!MDStatistic methodsFor: 'accessing' stamp: 'pmm 7/23/2007 22:49'!
version
	^version! !

!MDStatistic methodsFor: 'accessing' stamp: 'pmm 7/23/2007 22:53'!
version: aString 
	version := aString! !

Error subclass: #MDError
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Memcached-Exceptions'!

MDError subclass: #MDArgumentError
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Memcached-Exceptions'!

MDError subclass: #MDClientError
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Memcached-Exceptions'!

!MDClientError commentStamp: 'pmm 7/22/2007 10:28' prior: 0!
Means some sort of client error in the input line, i.e. the input doesn't conform to the protocol in some way.
!

MDError subclass: #MDConnectionClosedError
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Memcached-Exceptions'!

MDError subclass: #MDKeyNotFoundError
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Memcached-Exceptions'!

MDError subclass: #MDNoSuchCommandError
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Memcached-Exceptions'!

!MDNoSuchCommandError commentStamp: 'pmm 7/22/2007 10:28' prior: 0!
Means the client sent a nonexistent command name.
!

MDError subclass: #MDProtocolError
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Memcached-Exceptions'!

!MDProtocolError commentStamp: 'pmm 7/22/2007 10:48' prior: 0!
Means some sort of server error in the output line, i.e. the output doesn't conform to the protocol in some way.!

MDError subclass: #MDServerError
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Memcached-Exceptions'!

!MDServerError commentStamp: 'pmm 7/22/2007 10:30' prior: 0!
Means some sort of server error prevents the server from carrying out the command. In cases of severe server errors, which make it impossible to continue serving the client (this shouldn't normally happen), the server will close the connection after sending the error line. This is the only case in which the server closes a connection to a client.!
MDUtf8Codec initialize!
